from ..crazy import CrazyDragon

from threading import Thread

from .._base._controller_base.integral_loop import _dot_thrust
from .._base._controller_base.integral_loop import _thrust_clip

from .._base._controller_base.optimus_prime import _command_as_RPY
from .._base._controller_base.optimus_prime import _command_is_not_in_there

from .._base._controller_base.constants import alpha

from numpy import zeros, array

from time import sleep



class Controller( Thread ):

    def __init__( self, cf: CrazyDragon, config ):

        super().__init__()

        self.daemon = True

        self.cf = cf
        self.dt = config['dt']
        self.n  = config['n']

        self.acc_cmd = zeros(3)
        self.yaw_cmd = zeros(1)
        self.command = zeros(4)
        self.thrust  = array( [alpha * 9.81], dtype=int )


    def init_send_setpoint( self ):
        ## commander
        commander = self.cf.commander
        ## initialize
        commander.send_setpoint( 0, 0, 0, 0 )
        self.cf.ready_for_command = True


    def stop_send_setpoint( self ):
        ## commander
        commander = self.cf.commander
        ## stop command
        self.cf.command[:] = zeros(3)
        ## stop signal
        self.cf.ready_for_command = False

        for _ in range( 50 ):
            commander.send_setpoint( 0, 0, 0, 10001 )
            sleep( 0.02 )

        commander.send_stop_setpoint()


    def run( self ):

        cf        = self.cf
        commander = cf.commander

        n  = self.n
        dt = self.dt / n

        att_cur = cf.att
        acc_cur = cf.acc

        acc_cmd = self.acc_cmd
        yaw_cmd = self.yaw_cmd
        command = self.command
        thrust  = self.thrust

        while not cf.ready_for_command:
            sleep( 0.1 )

        print( 'controller starts working' )

        while cf.ready_for_command:

            acc_cmd[:] = cf.command
            yaw_cmd[:] = cf.yaw_cmd

            _command_is_not_in_there( att_cur, acc_cmd )

            _command_as_RPY( acc_cmd, command )

            if ( acc_cmd[2] == 0 ):
                sleep( dt )

            else:
                for _ in range( n ):

                    ## Thrust I loop
                    thrust[0] += _dot_thrust( command, acc_cur )

                    thrust[0] = _thrust_clip( thrust[0] )

                    ## YawRate PD loop
                    command[2] = 4.0 * ( yaw_cmd - att_cur[2] )

                    if   ( command[2] >  120 ): command[2] =  120
                    elif ( command[2] < -120 ): command[2] = -120

                    commander.send_setpoint(
                        command[0],
                        command[1],
                        -command[2],
                        thrust[0]
                    )

                    sleep( dt )

        print( 'controller end' )


from numpy.linalg import norm

from .constants import *



K_T = alpha * Kp


def _dot_thrust( command, acc_cur ):
    ## difference
    dot_acc = command[3] - norm( acc_cur )
    ## command
    dot_thr = K_T * dot_acc
    ## return
    return dot_thr


def _thrust_clip( thrust_cmd ):
    ## thrust clip thrust
    if ( thrust_cmd > 60000 ):
        thrust_cmd = 60000
    elif ( thrust_cmd < 10001 ):
        thrust_cmd = 10001

    return int( thrust_cmd )


import matplotlib.pyplot as plt

from numpy import arange



def plot_T( acc, acccmd, vel, pos, poscmd, date, idx1, idx2, dt=0.1 ):

    tim = arange( 0, len(acc[0,:]) ) * dt

    fig = plt.figure( figsize=(14,14) )

    ax1 = fig.add_subplot( 331 )
    ax2 = fig.add_subplot( 332 )
    ax3 = fig.add_subplot( 333 )
    ax4 = fig.add_subplot( 334 )
    ax5 = fig.add_subplot( 335 )
    ax6 = fig.add_subplot( 336 )
    ax7 = fig.add_subplot( 337 )
    ax8 = fig.add_subplot( 338 )
    ax9 = fig.add_subplot( 339 )

    ax1.plot( tim, acc[0,:]   , label='acc x' )
    ax2.plot( tim, acc[1,:]   , label='acc y' )
    ax3.plot( tim, acc[2,:]   , label='acc z' )
    ax1.plot( tim, acccmd[0,:], label='acc command x' )
    ax2.plot( tim, acccmd[1,:], label='acc command y' )
    ax3.plot( tim, acccmd[2,:], label='acc command z' )

    ax4.plot( tim, vel[0,:], label='velocity x' )
    ax5.plot( tim, vel[1,:], label='velocity y' )
    ax6.plot( tim, vel[2,:], label='velocity z' )

    ax7.plot( tim, pos[0,:]   , label='position x' )
    ax8.plot( tim, pos[1,:]   , label='position y' )
    ax9.plot( tim, pos[2,:]   , label='position z' )
    ax7.plot( tim, poscmd[0,:], label='position command x' )
    ax8.plot( tim, poscmd[1,:], label='position command y' )
    ax9.plot( tim, poscmd[2,:], label='position command z' )

    convenience( ax1 )
    convenience( ax2 )
    convenience( ax3 )
    convenience( ax4 )
    convenience( ax5 )
    convenience( ax6 )
    convenience( ax7 )
    convenience( ax8 )
    convenience( ax9 )

    axvline( ax1, tim, idx1, idx2 )
    axvline( ax2, tim, idx1, idx2 )
    axvline( ax3, tim, idx1, idx2 )
    axvline( ax4, tim, idx1, idx2 )
    axvline( ax5, tim, idx1, idx2 )
    axvline( ax6, tim, idx1, idx2 )
    axvline( ax7, tim, idx1, idx2 )
    axvline( ax8, tim, idx1, idx2 )
    axvline( ax9, tim, idx1, idx2 )

    plt.savefig( f'./flight_data/{date}/T.png' )


def plot_Thrust( thrust, thrustcmd, date, idx1, idx2, dt=0.1 ):

    tim = arange( 0, len(thrust[0,:]) ) * dt

    fig = plt.figure( figsize=(7,7) )

    ax1 = fig.add_subplot(111)
    ax1.plot( tim, thrust[0,:]   , label='thrust' )
    ax1.plot( tim, thrustcmd[0,:], label='thrust command' )

    convenience( ax1 )

    axvline( ax1, tim, idx1, idx2 )

    plt.savefig( f'./flight_data/{date}/thrust.png' )


def plot_R( att, attcmd, date, idx1, idx2, dt=0.1 ):

    tim = arange( 0, len(att[0,:]) ) * dt

    fig = plt.figure( figsize=(14,14) )

    ax1 = fig.add_subplot( 311 )
    ax2 = fig.add_subplot( 312 )
    ax3 = fig.add_subplot( 313 )

    ax1.plot( tim, att[0,:]   , label='euler att x' )
    ax2.plot( tim, att[1,:]   , label='euler att y' )
    ax3.plot( tim, att[2,:]   , label='euler att z' )
    ax1.plot( tim, attcmd[0,:], label='euler att command x' )
    ax2.plot( tim, attcmd[1,:], label='euler att command y' )
    ax3.plot( tim, attcmd[2,:], label='euler att command z' )

    convenience( ax1 )
    convenience( ax2 )
    convenience( ax3 )

    axvline( ax1, tim, idx1, idx2 )
    axvline( ax2, tim, idx1, idx2 )
    axvline( ax3, tim, idx1, idx2 )

    plt.savefig( f'./flight_data/{date}/R.png' )




def convenience( axn: plt.Axes ):
    axn.tick_params( axis='both', labelsize=7 )
    axn.legend( fontsize=8 )
    axn.grid()


def axvline( axn: plt.Axes, tim, idx1: int, idx2: int ):

    if ( ( idx1 == -1 ) or ( idx2 == -1 ) ):
        return

    axn.axvline( tim[idx1], tim[idx2], color='g', alpha=0.3, label='killer mode' )





def smooth_command(des, cur, t, T=1):
    if t < T/2:
        return C1(des, cur, t, T)
    elif t < T:
        return C2(des, cur, t, T)
    else:
        return des


def C1(des, cur, t, T):
    out = (2/T**2) * (des - cur) * t**2 + cur

    return out


def C2(des, cur, t, T):
    out = (2/T**2) * (cur - des) * (t-T)**2 + des

    return out


__version__ = "5.1.0"

from .control import Controller

from .crazy import CrazyDragon

from .guidance import Guidance

from .guidance import utils

from .navigation import Navigation

from .recorder import Recorder



from threading import Thread

from ..crazy import CrazyDragon

from ..control import Controller
from ..control import alpha

from .visualizer import plot_R, plot_T, plot_Thrust

from os import system

import datetime

from numpy import zeros, save

from time import sleep



class Recorder( Thread ):

    def __init__( self , _cf: CrazyDragon, CTR: Controller, n=18000 ):

        super().__init__()

        self._cf = _cf
        self.CTR = CTR

        self.recording = True

        self.G_Start = -1
        self.G_Stopd = -1

        self.idxn = 0

        self.record_datas = {
            'acc'   : zeros((3,n)),
            'acccmd': zeros((3,n)),
            'vel'   : zeros((3,n)),
            'pos'   : zeros((3,n)),
            'poscmd': zeros((3,n)),
            'att'   : zeros((3,n)),
            'cmd'   : zeros((4,n)),
            'thrust': zeros((1,n))
        }

        self.data_pointer = {
            'acc'   : _cf.acc,
            'acccmd': _cf.command,
            'vel'   : _cf.vel,
            'pos'   : _cf.pos,
            'poscmd': _cf.des,
            'att'   : _cf.att,
            'cmd'   : CTR.command,
            'thrust': CTR.thrust
        }


    def run( self ):

        data_pointer = self.data_pointer
        record_datas = self.record_datas

        while self.recording:

            for key, pointer in data_pointer.items():

                record_datas[key][:,self.idxn] = pointer[:]

            self.idxn += 1

            sleep( 0.1 )

    
    def join( self ):

        self.recording = False

        super().join()

        acc    = self.record_datas['acc']
        acccmd = self.record_datas['acccmd']
        vel    = self.record_datas['vel']
        pos    = self.record_datas['pos']
        poscmd = self.record_datas['poscmd']
        att    = self.record_datas['att']
        cmd    = self.record_datas['cmd']
        thrust = self.record_datas['thrust']

        attcmd    = cmd[0:3,:]
        thrustcmd = cmd[ 3 ,:] * alpha

        d = datetime.datetime.now()
        date      = f'{d.year}-{d.month}-{d.day:02}-{d.hour:02}-{d.minute:02}-{d.second:02}'
        self.date = date

        system( f'cd ./flight_data && mkdir {date}' )

        save( f'./flight_data/{date}/acc_cmd', acccmd )
        save( f'./flight_data/{date}/acc'    , acc )
        save( f'./flight_data/{date}/vel'    , vel )
        save( f'./flight_data/{date}/pos_cmd', poscmd )
        save( f'./flight_data/{date}/pos'    , pos )
        save( f'./flight_data/{date}/command', cmd )
        save( f'./flight_data/{date}/att_cmd', attcmd )
        save( f'./flight_data/{date}/att'    , att )

        plot_T( acc, acccmd, vel, pos, date, self.G_Start, self.G_Stopd )
        plot_R( att, attcmd, date, self.G_Start, self.G_Stopd )
        plot_Thrust( thrust, thrustcmd, self.G_Start, self.G_Stopd )